// Craig S. Kaplan 
// CS 106 Winter 2018 Assignment 04

// The "Piece" class is a way to represent the information
// about a single Tangram piece.  An object of type Piece 
// holds an array of PVectors describing the outline of one
// piece, and includes methods to help you move it around
// and rotate it.
class Piece
{
  // An array of points.  You should never have to use this 
  // array directly -- let the Piece manage it for you by sending
  // it messages (i.e., calling its methods).
  PVector[] pts;
  
  // The constructor.  You'll never need to use this directly
  // (that is, don't ever say "new Piece( ... )").
  Piece( float[] coords )
  {
    pts = new PVector[ coords.length / 2 ];
    for( int idx = 0; idx < pts.length; ++idx ) {
      pts[idx] = new PVector( 
        map( coords[2*idx], 0, 1, 200, 400 ),
        map( coords[2*idx+1], 0, 1, 200, 400 ) );
    }
  }
  
  // Ask a Piece to draw itself to the sketch window.  This
  // method does the beginShape()/vertex()/endShape()
  // stuff, and assumes that you've already set the fill
  // and stroke properties that you want.
  void drawPiece()
  {
    beginShape();
    for( int idx = 0; idx < pts.length; ++idx ) {
      vertex( pts[idx].x, pts[idx].y );
    }
    endShape( CLOSE );
  }
  
  // Ask a Piece to move to a new position.  The two numbers
  // tell the Piece how far to move in the sketch window.  For
  // example, p.movePiece( 3, 10 ) would move the Piece p 
  // 3 pixels to the right and 10 pixels down.
  void movePiece( float dx, float dy )
  {
    for( int idx = 0; idx < pts.length; ++idx ) {
      pts[idx].x += dx;
      pts[idx].y += dy;
    }
  }
  
  // Ask a Piece to rotate itself clockwise by 45 degrees around
  // its centre.
  void rotatePiece()
  {
    // The math here is a bit complicated and won't be explained.
    // Basically, there's a way to compute the place to rotate 
    // around, and a way to rotate a point by 45 degrees.  Happily,
    // you don't need to know how the math works in order to use
    // the method.
    
    float cx = 0.0;
    float cy = 0.0;
    
    for( int idx = 0; idx < pts.length; ++idx ) {
      cx += pts[idx].x;
      cy += pts[idx].y;
    }
    
    cx /= pts.length;
    cy /= pts.length;
    float r2 = sqrt( 2.0 ) * 0.5;
    
    for( int idx = 0; idx < pts.length; ++idx ) {
      float vx = pts[idx].x - cx;
      float vy = pts[idx].y - cy;
      pts[idx].x = cx + r2 * vx - r2 * vy;
      pts[idx].y = cy + r2 * vx + r2 * vy;
    }
  }
  
  // Ask a piece if the given x and y coordinates are contained
  // within it.  In other words, perform a hit test against the
  // shape of this piece.  Return a boolean value that tells
  // you whether a hit occurred.
  boolean hitTest( float x, float y )
  {
    // Again, don't worry about how this works.  It's actually
    // a bit tricky, and uses an even trickier helper function
    // defined below.
    
    if( pts.length == 4 ) {
      return hitTestTriangle( x, y, pts[0], pts[1], pts[3] ) ||
        hitTestTriangle( x, y, pts[1], pts[2], pts[3] );
    } else {
      return hitTestTriangle( x, y, pts[0], pts[1], pts[2] );
    }
  }
}

// A master function that builds an array holding a complete Tangram
// set, arranged into a square.  This is the main function you should
// use to get the Tangram when the sketch first starts, and when
// the user presses the "Reset" button.  Call this function, and put
// the result into a global variable.
Piece[] initPieces()
{
  Piece[] ret = new Piece[ tangram_coords.length ];
  
  for( int idx = 0; idx < ret.length; ++idx ) {
    ret[idx] = new Piece( tangram_coords[idx] );
  }
  
  return ret;
}

///////////////
//
// You don't need to read anything below this point if you don't want to.
// It's internal pieces of data and functions to support the main stuff
// above.
//
///////////////

// x and y coordinates of the vertices of all the Tangram pieces, as
// an array of arrays of floats.
float[][] tangram_coords = {
  { 0, 0, 0.25, 0.25, 0.5, 0 },
  { 0.5, 0, 1, 0.5, 1, 0 },
  { 0.25, 0.25, 0.5, 0.5, 0.75, 0.25, 0.5, 0 },
  { 0.5, 0.5, 0.75, 0.75, 0.75, 0.25 },
  { 0.75, 0.25, 0.75, 0.75, 1.0, 1.0, 1.0, 0.5 },
  { 0, 0, 0, 1, 0.5, 0.5 },
  { 0.5, 0.5, 0, 1, 1, 1 } };

// 2D cross product.
float det( PVector v, PVector w )
{
  return v.x * w.y - v.y * w.x;
}

// Test if the point (x,y) is in the Triangle ABC.
boolean hitTestTriangle( float x, float y, PVector A, PVector B, PVector C )
{
  // Algorithm adapted from http://mathworld.wolfram.com/TriangleInterior.html
  
  PVector v = new PVector( x, y ); // PVector.sub( new PVector( x, y ), A );
  PVector v1 = PVector.sub( B, A );
  PVector v2 = PVector.sub( C, A );
  float a = (det(v,v2) - det(A,v2)) / det(v1,v2);
  float b = (det(A,v1) - det(v,v1)) / det(v1,v2);
  
  return (a>=0) && (b>=0) && ((a+b)<=1);
}